# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from trytond.exceptions import UserError
from trytond.model.exceptions import ValidationError


class WrongDateofBirth(ValidationError):
    pass


class DateHealedBeforeDx(ValidationError):
    pass


class EndTreatmentDateBeforeStart(ValidationError):
    pass


class MedEndDateBeforeStart(ValidationError):
    pass


class NextDoseBeforeFirst(ValidationError):
    pass


class DrugPregnancySafetyCheck(ValidationError):
    pass


class NoAssociatedHealthProfessional(ValidationError):
    pass


class EvaluationEndBeforeStart(ValidationError):
    pass


class MustBeAPerson(ValidationError):
    pass


class DupOfficialName(ValidationError):
    pass


class FedAccountMismatch(ValidationError):
    pass


class BirthCertDateMismatch(ValidationError):
    pass


class NoAppointmentSelected(UserError):
    pass


class CanNotModifyVaccination(UserError):
    pass



# -*- coding: utf-8 -*-
##############################################################################
#
#    GNU Health: The Free Health and Hospital Information System
#    Copyright (C) 2008-2022 Luis Falcon <lfalcon@gnusolidario.org>
#    Copyright (C) 2011-2022 GNU Solidario <health@gnusolidario.org>
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from trytond.pool import Pool
from . import health
from . import core
from . import sequences
from . import wizard
from . import report


def register():
    Pool.register(
        health.OperationalArea,
        health.OperationalSector,
        health.DomiciliaryUnit,
        health.FederationCountryConfig,
        health.Occupation,
        health.Ethnicity,
        health.BirthCertificate,
        health.DeathCertificate,
        health.Party,
        health.ContactMechanism,
        health.PersonName,
        health.PartyAddress,
        health.DrugDoseUnits,
        health.MedicationFrequency,
        health.DrugForm,
        health.DrugRoute,
        health.MedicalSpecialty,
        health.HealthInstitution,
        health.HealthInstitutionSpecialties,
        health.HealthInstitutionOperationalSector,
        health.HealthInstitutionO2M,
        health.HospitalBuilding,
        health.HospitalUnit,
        health.HospitalOR,
        health.HospitalWard,
        health.HospitalBed,
        health.HealthProfessional,
        health.HealthProfessionalSpecialties,
        health.PhysicianSP,
        health.Family,
        health.FamilyMember,
        health.MedicamentCategory,
        health.Medicament,
        health.ImmunizationSchedule,
        health.ImmunizationScheduleLine,
        health.ImmunizationScheduleDose,
        health.PathologyCategory,
        health.PathologyGroup,
        health.Pathology,
        health.DiseaseMembers,
        health.BirthCertExtraInfo,
        health.DeathCertExtraInfo,
        health.DeathUnderlyingCondition,
        health.ProcedureCode,
        health.InsurancePlan,
        health.Insurance,
        health.AlternativePersonID,
        health.Product,
        health.PatientData,
        health.PatientDiseaseInfo,
        health.Appointment,
        health.AppointmentReport,
        health.OpenAppointmentReportStart,
        health.PatientPrescriptionOrder,
        health.PrescriptionLine,
        health.PatientMedication,
        health.PatientVaccination,
        health.PatientEvaluation,
        health.Directions,
        health.SecondaryCondition,
        health.DiagnosticHypothesis,
        health.SignsAndSymptoms,
        health.PatientECG,
        health.ProductTemplate,
        health.PageOfLife,
        health.Commands,
        health.Modules,
        health.Help,
        wizard.wizard_check_immunization_status.CheckImmunizationStatusInit,
        sequences.GnuHealthSequences,
        sequences.PatientSequence,
        sequences.PatientEvaluationSequence,
        sequences.AppointmentSequence,
        sequences.PrescriptionSequence,
        module='health', type_='model')

    Pool.register(
        health.OpenAppointmentReport,
        wizard.wizard_appointment_evaluation.CreateAppointmentEvaluation,
        wizard.wizard_check_immunization_status.CheckImmunizationStatus,
        module='health', type_='wizard')
    Pool.register(
        report.health_report.PatientDiseaseReport,
        report.health_report.PatientMedicationReport,
        report.health_report.PatientVaccinationReport,
        report.immunization_status_report.ImmunizationStatusReport,
        module='health', type_='report')



##############################################################################
#
#    GNU Health HMIS: The Free Health and Hospital Information System
#    Copyright (C) 2008-2022 Luis Falcon <falcon@gnuhealth.org>
#    Copyright (C) 2011-2022 GNU Solidario <health@gnusolidario.org>
#
#    The GNU Health HMIS component is part of the GNU Health project
#    www.gnuhealth.org
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

# Core, commonly used ojects
import pytz

from dateutil.relativedelta import relativedelta
from datetime import datetime
from trytond.transaction import Transaction
from trytond.pool import Pool
from trytond.i18n import gettext

from .exceptions import (NoAssociatedHealthProfessional)


def convert_date_timezone(sdate, target):
    """
    Convert dates from UTC to local timezone and viceversa
    Datetime values are stored in UTC, so we need conversion
    """

    Company = Pool().get('company.company')

    institution_timezone = None
    company_id = Transaction().context.get('company')
    if company_id:
        company = Company(company_id)
        if company.timezone:
            institution_timezone = pytz.timezone(company.timezone)

    if (target == 'utc'):
        # Convert date to UTC timezone
        res = institution_timezone.localize(sdate).astimezone(pytz.utc)
    else:
        # Convert from UTC to institution local timezone
        res = pytz.utc.localize(sdate).astimezone(institution_timezone)
    return res


def compute_age_from_dates(dob, deceased, dod, gender, caller, extra_date):
    """ Get the person's age.

    Calculate the current age of the patient or age at time of death.

    Returns:
    If caller == 'age': str in Y-M-D,
       caller == 'childbearing_age': boolean,
       caller == 'raw_age': [Y, M, D]

    """
    today = datetime.today().date()

    if dob:
        start = datetime.strptime(str(dob), '%Y-%m-%d')
        end = datetime.strptime(str(today), '%Y-%m-%d')

        if extra_date:
            end = datetime.strptime(str(extra_date), '%Y-%m-%d')

        if deceased and dod:
            end = datetime.strptime(
                        str(dod), '%Y-%m-%d %H:%M:%S')

        rdelta = relativedelta(end, start)

        years_months_days = str(rdelta.years) + 'y ' \
            + str(rdelta.months) + 'm ' \
            + str(rdelta.days) + 'd'

    else:
        return None

    if caller == 'age':
        return years_months_days

    elif caller == 'childbearing_age':
        if (rdelta.years >= 11
           and rdelta.years <= 55 and gender == 'f'):
            return True
        else:
            return False

    elif caller == 'raw_age':
        return [rdelta.years, rdelta.months, rdelta.days]

    else:
        return None


def get_institution():
    # Retrieve the institution associated to this GNU Health instance
    # That is associated to the Company.
    company = Transaction().context.get('company')

    cursor = Transaction().connection.cursor()
    cursor.execute('SELECT party FROM company_company WHERE id=%s \
        LIMIT 1', (company,))
    party_id = cursor.fetchone()
    if party_id:
        cursor = Transaction().connection.cursor()
        cursor.execute('SELECT id FROM gnuhealth_institution WHERE \
            name = %s LIMIT 1', (party_id[0],))
        institution_id = cursor.fetchone()
        if (institution_id):
            return int(institution_id[0])


def get_health_professional(required=True):
    # Get the professional associated to the internal user id
    # that logs into GNU Health
    # If the method is called with the arg "required" as False, then
    # the error message won't be shown in the case of not finding
    # the corresponding healthprof (eg, creating a new appointment)
    cursor = Transaction().connection.cursor()
    User = Pool().get('res.user')
    user = User(Transaction().user)
    login_user_id = int(user.id)
    cursor.execute('SELECT id FROM party_party WHERE is_healthprof=True \
        AND internal_user = %s LIMIT 1', (login_user_id,))
    partner_id = cursor.fetchone()
    if partner_id:
        cursor = Transaction().connection.cursor()
        cursor.execute('SELECT id FROM gnuhealth_healthprofessional WHERE \
            name = %s LIMIT 1', (partner_id[0],))
        healthprof_id = cursor.fetchone()
        if (healthprof_id):
            return int(healthprof_id[0])
    else:
        if required:
            raise NoAssociatedHealthProfessional(gettext(
                ('health.msg_no_associated_health_professional'))
            )



# -*- coding: utf-8 -*-
##############################################################################
#
#    GNU Health: The Free Health and Hospital Information System
#    Copyright (C) 2008-2022 Luis Falcon <falcon@gnuhealth.org>
#    Copyright (C) 2011-2022 GNU Solidario <health@gnusolidario.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
from trytond.report import Report
from trytond.pool import Pool

__all__ = ['ImmunizationStatusReport']


class ImmunizationStatusReport(Report):
    __name__ = 'gnuhealth.immunization_status_report'

    @classmethod
    def get_context(cls, records, header, data):
        Sched = Pool().get('gnuhealth.immunization_schedule')
        Patient = Pool().get('gnuhealth.patient')
        patient = Patient(data['patient_id'])

        context = super(ImmunizationStatusReport, cls).get_context(
            records, header, data)

        context['patient'] = patient
        sched = Sched(data['immunization_schedule_id'])

        context['immunization_schedule'] = sched

        immunizations_to_check = \
            cls.get_immunizations_for_age(patient, sched)

        immunization_status = \
            cls.verify_status(immunizations_to_check)

        context['immunization_status'] = immunization_status

        return context

    @classmethod
    def get_immunizations_for_age(cls, patient, immunization_schedule):

        immunizations_for_age = []

        for vaccine in immunization_schedule.vaccines:

            for dose in vaccine.doses:
                dose_number, dose_age, age_unit = dose.dose_number, \
                    dose.age_dose, dose.age_unit

                p_age = [patient.age.split(' ')[0][:-1],
                         patient.age.split(' ')[1][:-1],
                         patient.age.split(' ')[2][:-1]]

                # Age of the person in years and months
                pyears, pmonths = int(p_age[0]), int(p_age[1])

                pmonths = (pyears*12)+pmonths

                if ((age_unit == 'months' and pmonths >= dose_age) or
                        (age_unit == 'years' and pyears >= dose_age)):
                    immunization_info = {
                        'patient': patient,
                        'vaccine': vaccine,
                        'dose': dose_number,
                        'dose_age': dose_age,
                        'age_unit': age_unit,
                        'status': None}

                    # Add to the list of this person immunization check
                    immunizations_for_age.append(immunization_info)

        return immunizations_for_age

    @classmethod
    def verify_status(cls, immunizations_to_check):
        Vaccination = Pool().get('gnuhealth.vaccination')

        result = []
        for immunization in immunizations_to_check:
            immunization['status'] = "missing"
            res = Vaccination.search_count([
                ('name', '=', immunization['patient']),
                ('dose', '=', immunization['dose']),
                ('vaccine.name', '=', immunization['vaccine'].vaccine.name),
                ])

            if res:
                immunization['status'] = 'ok'

            result.append(immunization)

        return result



##############################################################################
#
#    GNU Health HMIS: The Free Health and Hospital Information System
#    Copyright (C) 2008-2022 Luis Falcon <falcon@gnuhealth.org>
#    Copyright (C) 2011-2022 GNU Solidario <health@gnusolidario.org>
#
#    The GNU Health HMIS component is part of the GNU Health project
#    www.gnuhealth.org
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

# GNU Health HMIS sequences for this package

from trytond.model import (ModelView, ModelSingleton, ModelSQL,
                           ValueMixin, MultiValueMixin, fields)
from trytond import backend
from trytond.pyson import Id
from trytond.pool import Pool
from trytond.tools.multivalue import migrate_property

# Sequences
# The patient_sequence is no longer used
# The patient ID is taken from the person demographics
patient_sequence = fields.Many2One(
    'ir.sequence', 'Patient Sequence', required=True,
    domain=[('sequence_type', '=', Id(
        'health', 'seq_type_gnuhealth_patient'))])

patient_evaluation_sequence = fields.Many2One(
    'ir.sequence', 'Patient Evaluation Sequence', required=True,
    domain=[('sequence_type', '=', Id(
        'health', 'seq_type_gnuhealth_patient_evaluation'))])

appointment_sequence = fields.Many2One(
    'ir.sequence', 'Appointment Sequence', required=True,
    domain=[('sequence_type', '=', Id(
        'health', 'seq_type_gnuhealth_appointment'))])


prescription_sequence = fields.Many2One(
    'ir.sequence', 'Prescription Sequence', required=True,
    domain=[('sequence_type', '=', Id(
        'health', 'seq_type_gnuhealth_prescription'))])


# GNU HEALTH SEQUENCES
class GnuHealthSequences(ModelSingleton, ModelSQL, ModelView, MultiValueMixin):
    'Standard Sequences for GNU Health'
    __name__ = 'gnuhealth.sequences'

    patient_sequence = fields.MultiValue(
        patient_sequence)

    patient_evaluation_sequence = fields.MultiValue(
        patient_evaluation_sequence)

    appointment_sequence = fields.MultiValue(
        appointment_sequence)

    prescription_sequence = fields.MultiValue(
        prescription_sequence)

    @classmethod
    def default_patient_sequence(cls, **pattern):
        pool = Pool()
        ModelData = pool.get('ir.model.data')
        try:
            return ModelData.get_id('health',
                                    'seq_gnuhealth_patient')
        except KeyError:
            return None

    @classmethod
    def default_patient_evaluation_sequence(cls, **pattern):
        pool = Pool()
        ModelData = pool.get('ir.model.data')
        try:
            return ModelData.get_id('health',
                                    'seq_gnuhealth_patient_evaluation')
        except KeyError:
            return None

    @classmethod
    def default_appointment_sequence(cls, **pattern):
        pool = Pool()
        ModelData = pool.get('ir.model.data')
        try:
            return ModelData.get_id('health',
                                    'seq_gnuhealth_appointment')
        except KeyError:
            return None

    @classmethod
    def default_prescription_sequence(cls, **pattern):
        pool = Pool()
        ModelData = pool.get('ir.model.data')
        try:
            return ModelData.get_id('health',
                                    'seq_gnuhealth_prescription')
        except KeyError:
            return None


class _ConfigurationValue(ModelSQL):

    _configuration_value_field = None

    @classmethod
    def __register__(cls, module_name):
        exist = backend.TableHandler.table_exist(cls._table)

        super(_ConfigurationValue, cls).__register__(module_name)

        if not exist:
            cls._migrate_property([], [], [])

    @classmethod
    def _migrate_property(cls, field_names, value_names, fields):
        field_names.append(cls._configuration_value_field)
        value_names.append(cls._configuration_value_field)
        migrate_property(
            'gnuhealth.sequences', field_names, cls, value_names,
            fields=fields)


class PatientSequence(_ConfigurationValue, ModelSQL, ValueMixin):
    'Patient Sequences setup'
    __name__ = 'gnuhealth.sequences.patient_sequence'
    patient_sequence = patient_sequence
    _configuration_value_field = 'patient_sequence'

    @classmethod
    def check_xml_record(cls, records, values):
        return True


class PatientEvaluationSequence(_ConfigurationValue, ModelSQL, ValueMixin):
    'Patient Evaluation Sequence setup'
    __name__ = 'gnuhealth.sequences.patient_evaluation_sequence'
    patient_evaluation_sequence = patient_evaluation_sequence
    _configuration_value_field = 'patient_evaluation_sequence'

    @classmethod
    def check_xml_record(cls, records, values):
        return True


class AppointmentSequence(_ConfigurationValue, ModelSQL, ValueMixin):
    'Appointment Sequence setup'
    __name__ = 'gnuhealth.sequences.appointment_sequence'
    appointment_sequence = appointment_sequence
    _configuration_value_field = 'appointment_sequence'

    @classmethod
    def check_xml_record(cls, records, values):
        return True


class PrescriptionSequence(_ConfigurationValue, ModelSQL, ValueMixin):
    'Prescription Sequence setup'
    __name__ = 'gnuhealth.sequences.prescription_sequence'
    prescription_sequence = prescription_sequence
    _configuration_value_field = 'prescription_sequence'

    @classmethod
    def check_xml_record(cls, records, values):
        return True
